@@ -19,15 +19,12 @@ use uucore::os_str_as_bytes;
 use self::searcher::Searcher;
 use matcher::{ExactMatcher, Matcher, WhitespaceMatcher};
 use uucore::ranges::Range;
-use uucore::{format_usage, help_about, help_section, help_usage, show_error, show_if_err};
+use uucore::translate;
+use uucore::{format_usage, show_error, show_if_err};
 
 mod matcher;
 mod searcher;
 
-const USAGE: &str = help_usage!("cut.md");
-const ABOUT: &str = help_about!("cut.md");
-const AFTER_HELP: &str = help_section!("after help", "cut.md");
-
 struct Options<'a> {
     out_delimiter: Option<&'a [u8]>,
     line_ending: LineEnding,
@@ -107,7 +104,7 @@ fn cut_bytes<R: Read, W: Write>(
     Ok(())
 }
 
-// Output delimiter is explicitly specified
+/// Output delimiter is explicitly specified
 fn cut_fields_explicit_out_delim<R: Read, W: Write, M: Matcher>(
     reader: R,
     out: &mut W,
@@ -192,7 +189,7 @@ fn cut_fields_explicit_out_delim<R: Read, W: Write, M: Matcher>(
     Ok(())
 }
 
-// Output delimiter is the same as input delimiter
+/// Output delimiter is the same as input delimiter
 fn cut_fields_implicit_out_delim<R: Read, W: Write, M: Matcher>(
     reader: R,
     out: &mut W,
@@ -263,7 +260,7 @@ fn cut_fields_implicit_out_delim<R: Read, W: Write, M: Matcher>(
     Ok(())
 }
 
-// The input delimiter is identical to `newline_char`
+/// The input delimiter is identical to `newline_char`
 fn cut_fields_newline_char_delim<R: Read, W: Write>(
     reader: R,
     out: &mut W,
@@ -345,11 +342,11 @@ fn cut_fields<R: Read, W: Write>(
     }
 }
 
-fn cut_files(mut filenames: Vec<String>, mode: &Mode) {
+fn cut_files(mut filenames: Vec<OsString>, mode: &Mode) {
     let mut stdin_read = false;
 
     if filenames.is_empty() {
-        filenames.push("-".to_owned());
+        filenames.push(OsString::from("-"));
     }
 
     let mut out: Box<dyn Write> = if stdout().is_terminal() {
@@ -372,10 +369,14 @@ fn cut_files(mut filenames: Vec<String>, mode: &Mode) {
 
             stdin_read = true;
         } else {
-            let path = Path::new(&filename[..]);
+            let path = Path::new(filename);
 
             if path.is_dir() {
-                show_error!("{}: Is a directory", filename.maybe_quote());
+                show_error!(
+                    "{}: {}",
+                    filename.maybe_quote(),
+                    translate!("cut-error-is-directory")
+                );
                 set_exit_code(1);
                 continue;
             }
@@ -395,19 +396,22 @@ fn cut_files(mut filenames: Vec<String>, mode: &Mode) {
         }
     }
 
-    show_if_err!(out.flush().map_err_context(|| "write error".into()));
+    show_if_err!(
+        out.flush()
+            .map_err_context(|| translate!("cut-error-write-error"))
+    );
 }
 
-// Get delimiter and output delimiter from `-d`/`--delimiter` and `--output-delimiter` options respectively
-// Allow either delimiter to have a value that is neither UTF-8 nor ASCII to align with GNU behavior
-fn get_delimiters(matches: &ArgMatches) -> UResult<(Delimiter, Option<&[u8]>)> {
+/// Get delimiter and output delimiter from `-d`/`--delimiter` and `--output-delimiter` options respectively
+/// Allow either delimiter to have a value that is neither UTF-8 nor ASCII to align with GNU behavior
+fn get_delimiters(matches: &ArgMatches) -> UResult<(Delimiter<'_>, Option<&[u8]>)> {
     let whitespace_delimited = matches.get_flag(options::WHITESPACE_DELIMITED);
     let delim_opt = matches.get_one::<OsString>(options::DELIMITER);
     let delim = match delim_opt {
         Some(_) if whitespace_delimited => {
             return Err(USimpleError::new(
                 1,
-                "invalid input: Only one of --delimiter (-d) or -w option can be specified",
+                translate!("cut-error-delimiter-and-whitespace-conflict"),
             ));
         }
         Some(os_string) => {
@@ -423,11 +427,10 @@ fn get_delimiters(matches: &ArgMatches) -> UResult<(Delimiter, Option<&[u8]>)> {
                 {
                     return Err(USimpleError::new(
                         1,
-                        "the delimiter must be a single character",
+                        translate!("cut-error-delimiter-must-be-single-character"),
                     ));
-                } else {
-                    Delimiter::from(os_string)
                 }
+                Delimiter::from(os_string)
             }
         }
         None => {
@@ -461,6 +464,8 @@ mod options {
     pub const WHITESPACE_DELIMITED: &str = "whitespace-delimited";
     pub const COMPLEMENT: &str = "complement";
     pub const FILE: &str = "file";
+    // ignored option
+    pub const NOTHING: &str = "nothing";
 }
 
 #[uucore::main]
@@ -479,7 +484,7 @@ pub fn uumain(args: impl uucore::Args) -> UResult<()> {
         })
         .collect();
 
-    let matches = uu_app().try_get_matches_from(args)?;
+    let matches = uucore::clap_localization::handle_clap_result(uu_app(), args)?;
 
     let complement = matches.get_flag(options::COMPLEMENT);
     let only_delimited = matches.get_flag(options::ONLY_DELIMITED);
@@ -505,39 +510,50 @@ pub fn uumain(args: impl uucore::Args) -> UResult<()> {
         matches.get_one::<String>(options::CHARACTERS),
         matches.get_one::<String>(options::FIELDS),
     ) {
-        (1, Some(byte_ranges), None, None) => list_to_ranges(byte_ranges, complement).map(|ranges| {
-            Mode::Bytes(
-                ranges,
-                Options {
-                    out_delimiter,
-                    line_ending,
-                    field_opts: None,
-                },
-            )
-        }),
-        (1, None, Some(char_ranges), None) => list_to_ranges(char_ranges, complement).map(|ranges| {
-            Mode::Characters(
-                ranges,
-                Options {
-                    out_delimiter,
-                    line_ending,
-                    field_opts: None,
-                },
-            )
-        }),
-        (1, None, None, Some(field_ranges)) => list_to_ranges(field_ranges, complement).map(|ranges| {
-            Mode::Fields(
-                ranges,
-                Options {
-                    out_delimiter,
-                    line_ending,
-                    field_opts: Some(FieldOptions { delimiter, only_delimited })},
-            )
-        }),
-        (2.., _, _, _) => Err(
-            "invalid usage: expects no more than one of --fields (-f), --chars (-c) or --bytes (-b)".into()
-        ),
-        _ => Err("invalid usage: expects one of --fields (-f), --chars (-c) or --bytes (-b)".into()),
+        (1, Some(byte_ranges), None, None) => {
+            list_to_ranges(byte_ranges, complement).map(|ranges| {
+                Mode::Bytes(
+                    ranges,
+                    Options {
+                        out_delimiter,
+                        line_ending,
+                        field_opts: None,
+                    },
+                )
+            })
+        }
+
+        (1, None, Some(char_ranges), None) => {
+            list_to_ranges(char_ranges, complement).map(|ranges| {
+                Mode::Characters(
+                    ranges,
+                    Options {
+                        out_delimiter,
+                        line_ending,
+                        field_opts: None,
+                    },
+                )
+            })
+        }
+
+        (1, None, None, Some(field_ranges)) => {
+            list_to_ranges(field_ranges, complement).map(|ranges| {
+                Mode::Fields(
+                    ranges,
+                    Options {
+                        out_delimiter,
+                        line_ending,
+                        field_opts: Some(FieldOptions {
+                            delimiter,
+                            only_delimited,
+                        }),
+                    },
+                )
+            })
+        }
+
+        (2.., _, _, _) => Err(translate!("cut-error-multiple-mode-args")),
+        _ => Err(translate!("cut-error-missing-mode-arg")),
     };
 
     let mode_parse = match mode_parse {
@@ -546,24 +562,24 @@ pub fn uumain(args: impl uucore::Args) -> UResult<()> {
             Mode::Bytes(_, _) | Mode::Characters(_, _)
                 if matches.contains_id(options::DELIMITER) =>
             {
-                Err("invalid input: The '--delimiter' ('-d') option only usable if printing a sequence of fields".into())
+                Err(translate!("cut-error-delimiter-only-with-fields"))
             }
             Mode::Bytes(_, _) | Mode::Characters(_, _)
                 if matches.get_flag(options::WHITESPACE_DELIMITED) =>
             {
-                Err("invalid input: The '-w' option only usable if printing a sequence of fields".into())
+                Err(translate!("cut-error-whitespace-only-with-fields"))
             }
             Mode::Bytes(_, _) | Mode::Characters(_, _)
                 if matches.get_flag(options::ONLY_DELIMITED) =>
             {
-                Err("invalid input: The '--only-delimited' ('-s') option only usable if printing a sequence of fields".into())
+                Err(translate!("cut-error-only-delimited-only-with-fields"))
             }
             _ => Ok(mode),
         },
     };
 
-    let files: Vec<String> = matches
-        .get_many::<String>(options::FILE)
+    let files: Vec<OsString> = matches
+        .get_many::<OsString>(options::FILE)
         .unwrap_or_default()
         .cloned()
         .collect();
@@ -580,9 +596,10 @@ pub fn uumain(args: impl uucore::Args) -> UResult<()> {
 pub fn uu_app() -> Command {
     Command::new(uucore::util_name())
         .version(uucore::crate_version!())
-        .override_usage(format_usage(USAGE))
-        .about(ABOUT)
-        .after_help(AFTER_HELP)
+        .help_template(uucore::localized_help_template(uucore::util_name()))
+        .override_usage(format_usage(&translate!("cut-usage")))
+        .about(translate!("cut-about"))
+        .after_help(translate!("cut-after-help"))
         .infer_long_args(true)
         // While `args_override_self(true)` for some arguments, such as `-d`
         // and `--output-delimiter`, is consistent to the behavior of GNU cut,
@@ -596,7 +613,7 @@ pub fn uu_app() -> Command {
             Arg::new(options::BYTES)
                 .short('b')
                 .long(options::BYTES)
-                .help("filter byte columns from the input source")
+                .help(translate!("cut-help-bytes"))
                 .allow_hyphen_values(true)
                 .value_name("LIST")
                 .action(ArgAction::Append),
@@ -605,7 +622,7 @@ pub fn uu_app() -> Command {
             Arg::new(options::CHARACTERS)
                 .short('c')
                 .long(options::CHARACTERS)
-                .help("alias for character mode")
+                .help(translate!("cut-help-characters"))
                 .allow_hyphen_values(true)
                 .value_name("LIST")
                 .action(ArgAction::Append),
@@ -615,13 +632,13 @@ pub fn uu_app() -> Command {
                 .short('d')
                 .long(options::DELIMITER)
                 .value_parser(ValueParser::os_string())
-                .help("specify the delimiter character that separates fields in the input source. Defaults to Tab.")
+                .help(translate!("cut-help-delimiter"))
                 .value_name("DELIM"),
         )
         .arg(
             Arg::new(options::WHITESPACE_DELIMITED)
                 .short('w')
-                .help("Use any number of whitespace (Space, Tab) to separate fields in the input source (FreeBSD extension).")
+                .help(translate!("cut-help-whitespace-delimited"))
                 .value_name("WHITESPACE")
                 .action(ArgAction::SetTrue),
         )
@@ -629,7 +646,7 @@ pub fn uu_app() -> Command {
             Arg::new(options::FIELDS)
                 .short('f')
                 .long(options::FIELDS)
-                .help("filter field columns from the input source")
+                .help(translate!("cut-help-fields"))
                 .allow_hyphen_values(true)
                 .value_name("LIST")
                 .action(ArgAction::Append),
@@ -637,34 +654,41 @@ pub fn uu_app() -> Command {
         .arg(
             Arg::new(options::COMPLEMENT)
                 .long(options::COMPLEMENT)
-                .help("invert the filter - instead of displaying only the filtered columns, display all but those columns")
+                .help(translate!("cut-help-complement"))
                 .action(ArgAction::SetTrue),
         )
         .arg(
             Arg::new(options::ONLY_DELIMITED)
                 .short('s')
                 .long(options::ONLY_DELIMITED)
-                .help("in field mode, only print lines which contain the delimiter")
+                .help(translate!("cut-help-only-delimited"))
                 .action(ArgAction::SetTrue),
         )
         .arg(
             Arg::new(options::ZERO_TERMINATED)
                 .short('z')
                 .long(options::ZERO_TERMINATED)
-                .help("instead of filtering columns based on line, filter columns based on \\0 (NULL character)")
+                .help(translate!("cut-help-zero-terminated"))
                 .action(ArgAction::SetTrue),
         )
         .arg(
             Arg::new(options::OUTPUT_DELIMITER)
                 .long(options::OUTPUT_DELIMITER)
                 .value_parser(ValueParser::os_string())
-                .help("in field mode, replace the delimiter in output lines with this option's argument")
+                .help(translate!("cut-help-output-delimiter"))
                 .value_name("NEW_DELIM"),
         )
         .arg(
             Arg::new(options::FILE)
-            .hide(true)
-            .action(ArgAction::Append)
-            .value_hint(clap::ValueHint::FilePath)
+                .hide(true)
+                .action(ArgAction::Append)
+                .value_hint(clap::ValueHint::FilePath)
+                .value_parser(clap::value_parser!(OsString)),
+        )
+        .arg(
+            Arg::new(options::NOTHING)
+                .short('n')
+                .help("(ignored)")
+                .action(ArgAction::SetTrue),
         )
 }
