@@ -7,9 +7,9 @@
 
 // spell-checker:ignore DATETIME getmntinfo subsecond (fs) cifs smbfs
 
-#[cfg(any(target_os = "linux", target_os = "android"))]
+#[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
 const LINUX_MTAB: &str = "/etc/mtab";
-#[cfg(any(target_os = "linux", target_os = "android"))]
+#[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
 const LINUX_MOUNTINFO: &str = "/proc/self/mountinfo";
 #[cfg(all(unix, not(any(target_os = "aix", target_os = "redox"))))]
 static MOUNT_OPT_BIND: &str = "bind";
@@ -19,17 +19,18 @@ const MAX_PATH: usize = 266;
 static EXIT_ERR: i32 = 1;
 
 #[cfg(any(
-    windows,
     target_os = "freebsd",
     target_vendor = "apple",
     target_os = "netbsd",
     target_os = "openbsd"
 ))]
+use crate::os_str_from_bytes;
 #[cfg(windows)]
 use crate::show_warning;
 
-#[cfg(windows)]
 use std::ffi::OsStr;
+#[cfg(unix)]
+use std::os::unix::ffi::OsStrExt;
 #[cfg(windows)]
 use std::os::windows::ffi::OsStrExt;
 #[cfg(windows)]
@@ -59,19 +60,25 @@ fn to_nul_terminated_wide_string(s: impl AsRef<OsStr>) -> Vec<u16> {
 
 #[cfg(unix)]
 use libc::{
-    mode_t, strerror, S_IFBLK, S_IFCHR, S_IFDIR, S_IFIFO, S_IFLNK, S_IFMT, S_IFREG, S_IFSOCK,
+    S_IFBLK, S_IFCHR, S_IFDIR, S_IFIFO, S_IFLNK, S_IFMT, S_IFREG, S_IFSOCK, mode_t, strerror,
 };
-use std::borrow::Cow;
 #[cfg(unix)]
-use std::ffi::CStr;
-#[cfg(unix)]
-use std::ffi::CString;
+use std::ffi::{CStr, CString};
 use std::io::Error as IOError;
 #[cfg(unix)]
 use std::mem;
 #[cfg(windows)]
 use std::path::Path;
+use std::time::SystemTime;
+#[cfg(not(windows))]
 use std::time::UNIX_EPOCH;
+use std::{borrow::Cow, ffi::OsString};
+
+use std::fs::Metadata;
+#[cfg(unix)]
+use std::os::unix::fs::MetadataExt;
+#[cfg(unix)]
+use std::time::Duration;
 
 #[cfg(any(
     target_os = "linux",
@@ -87,7 +94,8 @@ pub use libc::statfs as StatFs;
     target_os = "dragonfly",
     target_os = "illumos",
     target_os = "solaris",
-    target_os = "redox"
+    target_os = "redox",
+    target_os = "cygwin",
 ))]
 pub use libc::statvfs as StatFs;
 
@@ -105,41 +113,106 @@ pub use libc::statfs as statfs_fn;
     target_os = "illumos",
     target_os = "solaris",
     target_os = "dragonfly",
-    target_os = "redox"
+    target_os = "redox",
+    target_os = "cygwin",
 ))]
 pub use libc::statvfs as statfs_fn;
 
-pub trait BirthTime {
-    fn birth(&self) -> Option<(u64, u32)>;
+#[derive(Debug, Copy, Clone)]
+pub enum MetadataTimeField {
+    Modification,
+    Access,
+    Change,
+    Birth,
 }
 
-use std::fs::Metadata;
-impl BirthTime for Metadata {
-    fn birth(&self) -> Option<(u64, u32)> {
-        self.created()
-            .ok()
-            .and_then(|t| t.duration_since(UNIX_EPOCH).ok())
-            .map(|e| (e.as_secs(), e.subsec_nanos()))
+impl From<&str> for MetadataTimeField {
+    /// Get a `MetadataTimeField` from a string, we expect the value
+    /// to come from clap, and be constrained there (e.g. if Modification is
+    /// not supported), and the default branch should not be reached.
+    fn from(value: &str) -> Self {
+        match value {
+            "ctime" | "status" => Self::Change,
+            "access" | "atime" | "use" => Self::Access,
+            "mtime" | "modification" => Self::Modification,
+            "birth" | "creation" => Self::Birth,
+            // below should never happen as clap already restricts the values.
+            _ => unreachable!("Invalid metadata time field."),
+        }
     }
 }
 
+#[cfg(unix)]
+fn metadata_get_change_time(md: &Metadata) -> Option<SystemTime> {
+    let mut st = UNIX_EPOCH;
+    let (secs, nsecs) = (md.ctime(), md.ctime_nsec());
+    if secs >= 0 {
+        st += Duration::from_secs(secs as u64);
+    } else {
+        st -= Duration::from_secs(-secs as u64);
+    }
+    if nsecs >= 0 {
+        st += Duration::from_nanos(nsecs as u64);
+    } else {
+        // Probably never the case, but cover just in case.
+        st -= Duration::from_nanos(-nsecs as u64);
+    }
+    Some(st)
+}
+
+#[cfg(not(unix))]
+fn metadata_get_change_time(_md: &Metadata) -> Option<SystemTime> {
+    // Not available.
+    None
+}
+
+pub fn metadata_get_time(md: &Metadata, md_time: MetadataTimeField) -> Option<SystemTime> {
+    match md_time {
+        MetadataTimeField::Change => metadata_get_change_time(md),
+        MetadataTimeField::Modification => md.modified().ok(),
+        MetadataTimeField::Access => md.accessed().ok(),
+        MetadataTimeField::Birth => md.created().ok(),
+    }
+}
+
+// TODO: Types for this struct are probably mostly wrong. Possibly, most of them
+// should be OsString.
 #[derive(Debug, Clone)]
 pub struct MountInfo {
     /// Stores `volume_name` in windows platform and `dev_id` in unix platform
     pub dev_id: String,
     pub dev_name: String,
     pub fs_type: String,
-    pub mount_root: String,
-    pub mount_dir: String,
+    pub mount_root: OsString,
+    pub mount_dir: OsString,
     /// We only care whether this field contains "bind"
     pub mount_option: String,
     pub remote: bool,
     pub dummy: bool,
 }
 
+#[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
+fn replace_special_chars(s: &[u8]) -> Vec<u8> {
+    use bstr::ByteSlice;
+
+    // Replace
+    //
+    // * ASCII space with a regular space character,
+    // * \011 ASCII horizontal tab with a tab character,
+    // * ASCII backslash with an actual backslash character.
+    //
+    s.replace(r"\040", " ")
+        .replace(r"\011", "	")
+        .replace(r"\134", r"\")
+}
+
 impl MountInfo {
-    #[cfg(any(target_os = "linux", target_os = "android"))]
-    fn new(file_name: &str, raw: &[&str]) -> Option<Self> {
+    #[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
+    fn new(file_name: &str, raw: &[&[u8]]) -> Option<Self> {
+        use std::ffi::OsStr;
+        use std::os::unix::ffi::OsStrExt;
+        use std::os::unix::ffi::OsStringExt;
+
         let dev_name;
         let fs_type;
         let mount_root;
@@ -152,24 +225,27 @@ impl MountInfo {
             // "man proc" for more details
             LINUX_MOUNTINFO => {
                 const FIELDS_OFFSET: usize = 6;
-                let after_fields = raw[FIELDS_OFFSET..].iter().position(|c| *c == "-").unwrap()
+                let after_fields = raw[FIELDS_OFFSET..]
+                    .iter()
+                    .position(|c| *c == b"-")
+                    .unwrap()
                     + FIELDS_OFFSET
                     + 1;
-                dev_name = raw[after_fields + 1].to_string();
-                fs_type = raw[after_fields].to_string();
-                mount_root = raw[3].to_string();
-                mount_dir = raw[4].to_string();
-                mount_option = raw[5].to_string();
+                dev_name = String::from_utf8_lossy(raw[after_fields + 1]).to_string();
+                fs_type = String::from_utf8_lossy(raw[after_fields]).to_string();
+                mount_root = OsStr::from_bytes(raw[3]).to_owned();
+                mount_dir = OsString::from_vec(replace_special_chars(raw[4]));
+                mount_option = String::from_utf8_lossy(raw[5]).to_string();
             }
             LINUX_MTAB => {
-                dev_name = raw[0].to_string();
-                fs_type = raw[2].to_string();
-                mount_root = String::new();
-                mount_dir = raw[1].to_string();
-                mount_option = raw[3].to_string();
+                dev_name = String::from_utf8_lossy(raw[0]).to_string();
+                fs_type = String::from_utf8_lossy(raw[2]).to_string();
+                mount_root = OsString::new();
+                mount_dir = OsString::from_vec(replace_special_chars(raw[1]));
+                mount_option = String::from_utf8_lossy(raw[3]).to_string();
             }
             _ => return None,
-        };
+        }
 
         let dev_id = mount_dev_id(&mount_dir);
         let dummy = is_dummy_filesystem(&fs_type, &mount_option);
@@ -220,6 +296,8 @@ impl MountInfo {
             // TODO: support the case when `GetLastError()` returns `ERROR_MORE_DATA`
             return None;
         }
+        // TODO: This should probably call `OsString::from_wide`, but unclear if
+        // terminating zeros need to be striped first.
         let mount_root = LPWSTR2String(&mount_root_buf);
 
         let mut fs_type_buf = [0u16; MAX_PATH];
@@ -250,8 +328,8 @@ impl MountInfo {
             dev_id: volume_name,
             dev_name,
             fs_type: fs_type.unwrap_or_default(),
-            mount_root,
-            mount_dir: String::new(),
+            mount_root: mount_root.into(), // TODO: We should figure out how to keep an OsString here.
+            mount_dir: OsString::new(),
             mount_option: String::new(),
             remote,
             dummy: false,
@@ -269,22 +347,21 @@ impl From<StatFs> for MountInfo {
     fn from(statfs: StatFs) -> Self {
         let dev_name = unsafe {
             // spell-checker:disable-next-line
-            CStr::from_ptr(&statfs.f_mntfromname[0])
+            CStr::from_ptr(statfs.f_mntfromname.as_ptr())
                 .to_string_lossy()
                 .into_owned()
         };
         let fs_type = unsafe {
             // spell-checker:disable-next-line
-            CStr::from_ptr(&statfs.f_fstypename[0])
+            CStr::from_ptr(statfs.f_fstypename.as_ptr())
                 .to_string_lossy()
                 .into_owned()
         };
-        let mount_dir = unsafe {
+        let mount_dir_bytes = unsafe {
             // spell-checker:disable-next-line
-            CStr::from_ptr(&statfs.f_mntonname[0])
-                .to_string_lossy()
-                .into_owned()
+            CStr::from_ptr(statfs.f_mntonname.as_ptr()).to_bytes()
         };
+        let mount_dir = os_str_from_bytes(mount_dir_bytes).unwrap().into_owned();
 
         let dev_id = mount_dev_id(&mount_dir);
         let dummy = is_dummy_filesystem(&fs_type, "");
@@ -295,7 +372,7 @@ impl From<StatFs> for MountInfo {
             dev_name,
             fs_type,
             mount_dir,
-            mount_root: String::new(),
+            mount_root: OsString::new(),
             mount_option: String::new(),
             remote,
             dummy,
@@ -303,7 +380,7 @@ impl From<StatFs> for MountInfo {
     }
 }
 
-#[cfg(all(unix, not(any(target_os = "aix", target_os = "redox"))))]
+#[cfg(all(unix, not(target_os = "redox")))]
 fn is_dummy_filesystem(fs_type: &str, mount_option: &str) -> bool {
     // spell-checker:disable
     match fs_type {
@@ -315,7 +392,11 @@ fn is_dummy_filesystem(fs_type: &str, mount_option: &str) -> bool {
         // for NetBSD 3.0
         | "kernfs"
         // for Irix 6.5
-        | "ignore" => true,
+        | "ignore"
+        // Linux initial root filesystem
+        | "rootfs"
+        // Binary format support pseudo-filesystem
+        | "binfmt_misc" => true,
         _ => fs_type == "none"
             && !mount_option.contains(MOUNT_OPT_BIND)
     }
@@ -330,7 +411,7 @@ fn is_remote_filesystem(dev_name: &str, fs_type: &str) -> bool {
 }
 
 #[cfg(all(unix, not(any(target_os = "aix", target_os = "redox"))))]
-fn mount_dev_id(mount_dir: &str) -> String {
+fn mount_dev_id(mount_dir: &OsStr) -> String {
     use std::os::unix::fs::MetadataExt;
 
     if let Ok(stat) = std::fs::metadata(mount_dir) {
@@ -341,40 +422,6 @@ fn mount_dev_id(mount_dir: &str) -> String {
     }
 }
 
-#[cfg(any(
-    target_os = "freebsd",
-    target_vendor = "apple",
-    target_os = "netbsd",
-    target_os = "openbsd"
-))]
-use libc::c_int;
-#[cfg(any(
-    target_os = "freebsd",
-    target_vendor = "apple",
-    target_os = "netbsd",
-    target_os = "openbsd"
-))]
-extern "C" {
-    #[cfg(all(target_vendor = "apple", target_arch = "x86_64"))]
-    #[link_name = "getmntinfo$INODE64"]
-    fn get_mount_info(mount_buffer_p: *mut *mut StatFs, flags: c_int) -> c_int;
-
-    #[cfg(any(
-        target_os = "netbsd",
-        target_os = "openbsd",
-        all(target_vendor = "apple", target_arch = "aarch64")
-    ))]
-    #[link_name = "getmntinfo"]
-    fn get_mount_info(mount_buffer_p: *mut *mut StatFs, flags: c_int) -> c_int;
-
-    // Rust on FreeBSD uses 11.x ABI for filesystem metadata syscalls.
-    // Call the right version of the symbol for getmntinfo() result to
-    // match libc StatFS layout.
-    #[cfg(target_os = "freebsd")]
-    #[link_name = "getmntinfo@FBSD_1.0"]
-    fn get_mount_info(mount_buffer_p: *mut *mut StatFs, flags: c_int) -> c_int;
-}
-
 use crate::error::UResult;
 #[cfg(any(
     target_os = "freebsd",
@@ -384,9 +431,9 @@ use crate::error::UResult;
     target_os = "windows"
 ))]
 use crate::error::USimpleError;
-#[cfg(any(target_os = "linux", target_os = "android"))]
+#[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
 use std::fs::File;
-#[cfg(any(target_os = "linux", target_os = "android"))]
+#[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
 use std::io::{BufRead, BufReader};
 #[cfg(any(
     target_vendor = "apple",
@@ -406,17 +453,17 @@ use std::slice;
 
 /// Read file system list.
 pub fn read_fs_list() -> UResult<Vec<MountInfo>> {
-    #[cfg(any(target_os = "linux", target_os = "android"))]
+    #[cfg(any(target_os = "linux", target_os = "android", target_os = "cygwin"))]
     {
         let (file_name, f) = File::open(LINUX_MOUNTINFO)
             .map(|f| (LINUX_MOUNTINFO, f))
             .or_else(|_| File::open(LINUX_MTAB).map(|f| (LINUX_MTAB, f)))?;
         let reader = BufReader::new(f);
         Ok(reader
-            .lines()
+            .split(b'\n')
             .map_while(Result::ok)
             .filter_map(|line| {
-                let raw_data = line.split_whitespace().collect::<Vec<&str>>();
+                let raw_data = line.split(|c| *c == b' ').collect::<Vec<&[u8]>>();
                 MountInfo::new(file_name, &raw_data)
             })
             .collect::<Vec<_>>())
@@ -429,9 +476,9 @@ pub fn read_fs_list() -> UResult<Vec<MountInfo>> {
     ))]
     {
         let mut mount_buffer_ptr: *mut StatFs = ptr::null_mut();
-        let len = unsafe { get_mount_info(&mut mount_buffer_ptr, 1_i32) };
+        let len = unsafe { libc::getmntinfo(&raw mut mount_buffer_ptr, 1_i32) };
         if len < 0 {
-            return Err(USimpleError::new(1, "get_mount_info() failed"));
+            return Err(USimpleError::new(1, "getmntinfo() failed"));
         }
         let mounts = unsafe { slice::from_raw_parts(mount_buffer_ptr, len as usize) };
         Ok(mounts
@@ -447,14 +494,14 @@ pub fn read_fs_list() -> UResult<Vec<MountInfo>> {
             unsafe { FindFirstVolumeW(volume_name_buf.as_mut_ptr(), volume_name_buf.len() as u32) };
         if INVALID_HANDLE_VALUE == find_handle {
             let os_err = IOError::last_os_error();
-            let msg = format!("FindFirstVolumeW failed: {}", os_err);
+            let msg = format!("FindFirstVolumeW failed: {os_err}");
             return Err(USimpleError::new(EXIT_ERR, msg));
         }
         let mut mounts = Vec::<MountInfo>::new();
         loop {
             let volume_name = LPWSTR2String(&volume_name_buf);
             if !volume_name.starts_with("\\\\?\\") || !volume_name.ends_with('\\') {
-                show_warning!("A bad path was skipped: {}", volume_name);
+                show_warning!("A bad path was skipped: {volume_name}");
                 continue;
             }
             if let Some(m) = MountInfo::new(volume_name) {
@@ -587,23 +634,15 @@ impl FsUsage {
         let mut number_of_free_clusters = 0;
         let mut total_number_of_clusters = 0;
 
-        let success = unsafe {
+        unsafe {
             let path = to_nul_terminated_wide_string(path);
             GetDiskFreeSpaceW(
                 path.as_ptr(),
-                &mut sectors_per_cluster,
-                &mut bytes_per_sector,
-                &mut number_of_free_clusters,
-                &mut total_number_of_clusters,
-            )
-        };
-        if 0 == success {
-            // Fails in case of CD for example
-            // crash!(
-            //     EXIT_ERR,
-            //     "GetDiskFreeSpaceW failed: {}",
-            //     IOError::last_os_error()
-            // );
+                &raw mut sectors_per_cluster,
+                &raw mut bytes_per_sector,
+                &raw mut number_of_free_clusters,
+                &raw mut total_number_of_clusters,
+            );
         }
 
         let bytes_per_cluster = sectors_per_cluster as u64 * bytes_per_sector as u64;
@@ -655,6 +694,7 @@ impl FsMeta for StatFs {
             not(target_os = "solaris"),
             not(target_os = "redox"),
             not(target_arch = "s390x"),
+            not(target_os = "cygwin"),
             target_pointer_width = "64"
         ))]
         return self.f_bsize;
@@ -663,6 +703,7 @@ impl FsMeta for StatFs {
             not(target_os = "freebsd"),
             not(target_os = "netbsd"),
             not(target_os = "redox"),
+            not(target_os = "cygwin"),
             any(
                 target_arch = "s390x",
                 target_vendor = "apple",
@@ -680,6 +721,7 @@ impl FsMeta for StatFs {
             target_os = "illumos",
             target_os = "solaris",
             target_os = "redox",
+            target_os = "cygwin",
             all(target_os = "android", target_pointer_width = "64"),
         ))]
         return self.f_bsize.try_into().unwrap();
@@ -807,9 +849,10 @@ impl FsMeta for StatFs {
         target_os = "openbsd"
     ))]
     fn fsid(&self) -> u64 {
-        let f_fsid: &[u32; 2] =
-            unsafe { &*(&self.f_fsid as *const nix::sys::statfs::fsid_t as *const [u32; 2]) };
-        (u64::from(f_fsid[0])) << 32 | u64::from(f_fsid[1])
+        // Use type inference to determine the type of f_fsid
+        // (libc::__fsid_t on Android, libc::fsid_t on other platforms)
+        let f_fsid: &[u32; 2] = unsafe { &*(&raw const self.f_fsid).cast() };
+        ((u64::from(f_fsid[0])) << 32) | u64::from(f_fsid[1])
     }
     #[cfg(not(any(
         target_vendor = "apple",
@@ -849,15 +892,17 @@ impl FsMeta for StatFs {
 }
 
 #[cfg(unix)]
-pub fn statfs<P>(path: P) -> Result<StatFs, String>
-where
-    P: Into<Vec<u8>>,
-{
-    match CString::new(path) {
+pub fn statfs(path: &OsStr) -> Result<StatFs, String> {
+    #[cfg(unix)]
+    let p = path.as_bytes();
+    #[cfg(not(unix))]
+    let p = path.into_string().unwrap();
+
+    match CString::new(p) {
         Ok(p) => {
             let mut buffer: StatFs = unsafe { mem::zeroed() };
             unsafe {
-                match statfs_fn(p.as_ptr(), &mut buffer) {
+                match statfs_fn(p.as_ptr(), &raw mut buffer) {
                     0 => Ok(buffer),
                     _ => {
                         let errno = IOError::last_os_error().raw_os_error().unwrap_or(0);
@@ -874,7 +919,7 @@ where
 }
 
 #[cfg(unix)]
-pub fn pretty_filetype<'a>(mode: mode_t, size: u64) -> &'a str {
+pub fn pretty_filetype(mode: mode_t, size: u64) -> String {
     match mode & S_IFMT {
         S_IFREG => {
             if size == 0 {
@@ -890,9 +935,9 @@ pub fn pretty_filetype<'a>(mode: mode_t, size: u64) -> &'a str {
         S_IFIFO => "fifo",
         S_IFSOCK => "socket",
         // TODO: Other file types
-        // See coreutils/gnulib/lib/file-type.c // spell-checker:disable-line
-        _ => "weird file",
+        _ => return format!("weird file ({:07o})", mode & S_IFMT),
     }
+    .to_owned()
 }
 
 pub fn pretty_fstype<'a>(fstype: i64) -> Cow<'a, str> {
@@ -915,6 +960,7 @@ pub fn pretty_fstype<'a>(fstype: i64) -> Cow<'a, str> {
         0x7372_7279 => "btrfs_test".into(),
         0x00C3_6400 => "ceph".into(),
         0x0027_E0EB => "cgroupfs".into(),
+        0x6367_7270 => "cgroup2fs".into(),
         0xFF53_4D42 => "cifs".into(),
         0x7375_7245 => "coda".into(),
         0x012F_F7B7 => "coh".into(),
@@ -1030,7 +1076,7 @@ mod tests {
         assert_eq!("character special file", pretty_filetype(S_IFCHR, 0));
         assert_eq!("regular file", pretty_filetype(S_IFREG, 1));
         assert_eq!("regular empty file", pretty_filetype(S_IFREG, 0));
-        assert_eq!("weird file", pretty_filetype(0, 0));
+        assert_eq!("weird file (0000000)", pretty_filetype(0, 0));
     }
 
     #[test]
@@ -1054,8 +1100,8 @@ mod tests {
         // spell-checker:ignore (word) relatime
         let info = MountInfo::new(
             LINUX_MOUNTINFO,
-            &"106 109 253:6 / /mnt rw,relatime - xfs /dev/fs0 rw"
-                .split_ascii_whitespace()
+            &b"106 109 253:6 / /mnt rw,relatime - xfs /dev/fs0 rw"
+                .split(|c| *c == b' ')
                 .collect::<Vec<_>>(),
         )
         .unwrap();
@@ -1069,8 +1115,8 @@ mod tests {
         // Test parsing with different amounts of optional fields.
         let info = MountInfo::new(
             LINUX_MOUNTINFO,
-            &"106 109 253:6 / /mnt rw,relatime master:1 - xfs /dev/fs0 rw"
-                .split_ascii_whitespace()
+            &b"106 109 253:6 / /mnt rw,relatime master:1 - xfs /dev/fs0 rw"
+                .split(|c| *c == b' ')
                 .collect::<Vec<_>>(),
         )
         .unwrap();
@@ -1080,8 +1126,8 @@ mod tests {
 
         let info = MountInfo::new(
             LINUX_MOUNTINFO,
-            &"106 109 253:6 / /mnt rw,relatime master:1 shared:2 - xfs /dev/fs0 rw"
-                .split_ascii_whitespace()
+            &b"106 109 253:6 / /mnt rw,relatime master:1 shared:2 - xfs /dev/fs0 rw"
+                .split(|c| *c == b' ')
                 .collect::<Vec<_>>(),
         )
         .unwrap();
@@ -1089,4 +1135,67 @@ mod tests {
         assert_eq!(info.fs_type, "xfs");
         assert_eq!(info.dev_name, "/dev/fs0");
     }
+
+    #[test]
+    #[cfg(any(target_os = "linux", target_os = "android"))]
+    fn test_mountinfo_dir_special_chars() {
+        let info = MountInfo::new(
+            LINUX_MOUNTINFO,
+            &br"317 61 7:0 / /mnt/f\134\040\011oo rw,relatime shared:641 - ext4 /dev/loop0 rw"
+                .split(|c| *c == b' ')
+                .collect::<Vec<_>>(),
+        )
+        .unwrap();
+
+        assert_eq!(info.mount_dir, r"/mnt/f\ 	oo");
+
+        let info = MountInfo::new(
+            LINUX_MTAB,
+            &br"/dev/loop0 /mnt/f\134\040\011oo ext4 rw,relatime 0 0"
+                .split(|c| *c == b' ')
+                .collect::<Vec<_>>(),
+        )
+        .unwrap();
+
+        assert_eq!(info.mount_dir, r"/mnt/f\ 	oo");
+    }
+
+    #[test]
+    #[cfg(any(target_os = "linux", target_os = "android"))]
+    fn test_mountinfo_dir_non_unicode() {
+        let info = MountInfo::new(
+            LINUX_MOUNTINFO,
+            &b"317 61 7:0 / /mnt/some-\xc0-dir-\xf3 rw,relatime shared:641 - ext4 /dev/loop0 rw"
+                .split(|c| *c == b' ')
+                .collect::<Vec<_>>(),
+        )
+        .unwrap();
+
+        assert_eq!(
+            info.mount_dir,
+            crate::os_str_from_bytes(b"/mnt/some-\xc0-dir-\xf3").unwrap()
+        );
+
+        let info = MountInfo::new(
+            LINUX_MOUNTINFO,
+            &b"317 61 7:0 / /mnt/some-\\040-dir-\xf3 rw,relatime shared:641 - ext4 /dev/loop0 rw"
+                .split(|c| *c == b' ')
+                .collect::<Vec<_>>(),
+        )
+        .unwrap();
+
+        // Note that the \040 above will have been substituted by a space.
+        assert_eq!(
+            info.mount_dir,
+            crate::os_str_from_bytes(b"/mnt/some- -dir-\xf3").unwrap()
+        );
+    }
+
+    #[test]
+    #[cfg(all(unix, not(target_os = "redox")))]
+    // spell-checker:ignore (word) binfmt
+    fn test_binfmt_misc_is_dummy() {
+        use super::is_dummy_filesystem;
+        assert!(is_dummy_filesystem("binfmt_misc", ""));
+    }
 }
