@@ -4,11 +4,12 @@
 // file that was distributed with this source code.
 // spell-checker:ignore (words) nosuchgroup groupname
 
+#[cfg(target_os = "linux")]
+use std::os::unix::ffi::OsStringExt;
 use uucore::process::getegid;
-use uutests::at_and_ucmd;
-use uutests::new_ucmd;
-use uutests::util::TestScenario;
-use uutests::util_name;
+use uutests::{at_and_ucmd, new_ucmd};
+#[cfg(not(target_vendor = "apple"))]
+use uutests::{util::TestScenario, util_name};
 
 #[test]
 fn test_invalid_option() {
@@ -370,7 +371,7 @@ fn test_traverse_symlinks() {
         (&["-P"][..], false, false),
         (&["-L"][..], true, true),
     ] {
-        let scenario = TestScenario::new("chgrp");
+        let scenario = TestScenario::new(util_name!());
 
         let (at, mut ucmd) = (scenario.fixtures.clone(), scenario.ucmd());
 
@@ -600,3 +601,65 @@ fn test_numeric_group_formats() {
     let final_gid = at.plus("test_file").metadata().unwrap().gid();
     assert_eq!(final_gid, first_group.as_raw());
 }
+
+#[test]
+#[cfg(target_os = "linux")]
+fn test_chgrp_non_utf8_paths() {
+    let (at, mut ucmd) = at_and_ucmd!();
+
+    let filename = std::ffi::OsString::from_vec(vec![0xFF, 0xFE]);
+    std::fs::write(at.plus(&filename), b"test content").unwrap();
+
+    // Get current user's primary group
+    let current_gid = getegid();
+
+    ucmd.arg(current_gid.to_string()).arg(&filename).succeeds();
+}
+
+#[test]
+fn test_chgrp_recursive_on_file() {
+    // Test for regression where `chgrp -R` on a regular file would fail
+    // with "Not a directory" error. This should succeed since there's nothing
+    // to recurse into, similar to GNU chgrp behavior.
+    // equivalent of tests/chgrp/recurse in GNU coreutils
+    use std::os::unix::fs::MetadataExt;
+    let (at, mut ucmd) = at_and_ucmd!();
+
+    at.touch("regular_file");
+
+    let current_gid = getegid();
+
+    ucmd.arg("-R")
+        .arg(current_gid.to_string())
+        .arg("regular_file")
+        .succeeds()
+        .no_stderr();
+
+    assert_eq!(
+        at.plus("regular_file").metadata().unwrap().gid(),
+        current_gid
+    );
+}
+
+#[test]
+fn test_chgrp_exit_code_not_being_overwritten_by_last_file() {
+    use std::os::unix::prelude::PermissionsExt;
+
+    let current_gid = getegid();
+    let (at, mut ucmd) = at_and_ucmd!();
+    at.mkdir("dir");
+    at.mkdir("dir/a");
+    at.mkdir("dir/b");
+    at.touch("dir/b/file");
+    at.touch("dir/a/file");
+    std::fs::set_permissions(at.plus("dir/a"), PermissionsExt::from_mode(0o0000)).unwrap();
+
+    // chgrp walks the dir alphabetically. Dir a does not have permissions so it fails, dir b does have
+    // permissions so it succeeds. We check that the overall command does fail although the
+    // last step succeeded.
+
+    ucmd.arg("-R")
+        .arg(current_gid.to_string())
+        .arg("dir")
+        .fails();
+}
