@@ -3,19 +3,30 @@
 // For the full copyright and license information, please view the LICENSE
 // file that was distributed with this source code.
 
-// spell-checker:ignore (ToDO) coreutil
+// spell-checker:ignore (ToDO) coreutil euid rgid
 
-use crate::common::util::{check_coreutil_version, expected_result, is_ci, whoami, TestScenario};
+use std::process::{Command, Stdio};
+use uutests::new_ucmd;
+use uutests::unwrap_or_return;
+use uutests::util::{TestScenario, check_coreutil_version, expected_result, is_ci, whoami};
+use uutests::util_name;
+
+#[cfg(all(feature = "chmod", feature = "chown"))]
+use tempfile::TempPath;
 
 const VERSION_MIN_MULTIPLE_USERS: &str = "8.31"; // this feature was introduced in GNU's coreutils 8.31
 
 #[test]
 fn test_invalid_arg() {
-    new_ucmd!().arg("--definitely-invalid").fails().code_is(1);
+    new_ucmd!().arg("--definitely-invalid").fails_with_code(1);
+}
+
+#[test]
+fn test_id_ignore() {
+    new_ucmd!().arg("-a").succeeds();
 }
 
 #[test]
-#[cfg(unix)]
 #[allow(unused_mut)]
 fn test_id_no_specified_user() {
     let ts = TestScenario::new(util_name!());
@@ -41,7 +52,6 @@ fn test_id_no_specified_user() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_single_user() {
     let test_users = [&whoami()[..]];
 
@@ -93,7 +103,6 @@ fn test_id_single_user() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_single_user_non_existing() {
     let args = &["hopefully_non_existing_username"];
     let ts = TestScenario::new(util_name!());
@@ -111,7 +120,6 @@ fn test_id_single_user_non_existing() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_name() {
     let ts = TestScenario::new(util_name!());
     for opt in ["--user", "--group", "--groups"] {
@@ -130,7 +138,6 @@ fn test_id_name() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_real() {
     let ts = TestScenario::new(util_name!());
     for opt in ["--user", "--group", "--groups"] {
@@ -145,27 +152,16 @@ fn test_id_real() {
 }
 
 #[test]
-#[cfg(all(unix, not(any(target_os = "linux", target_os = "android"))))]
+#[cfg(not(any(target_os = "linux", target_os = "android")))]
 fn test_id_pretty_print() {
     // `-p` is BSD only and not supported on GNU's `id`
     let username = whoami();
-
     let result = new_ucmd!().arg("-p").run();
-    if result.stdout_str().trim().is_empty() {
-        // this fails only on: "MinRustV (ubuntu-latest, feat_os_unix)"
-        // `rustc 1.40.0 (73528e339 2019-12-16)`
-        // run: /home/runner/work/coreutils/coreutils/target/debug/coreutils id -p
-        // thread 'test_id::test_id_pretty_print' panicked at 'Command was expected to succeed.
-        // stdout =
-        // stderr = ', tests/common/util.rs:157:13
-        println!("test skipped:");
-    } else {
-        result.success().stdout_contains(username);
-    }
+    result.success().stdout_contains(username);
 }
 
 #[test]
-#[cfg(all(unix, not(any(target_os = "linux", target_os = "android"))))]
+#[cfg(not(any(target_os = "linux", target_os = "android")))]
 fn test_id_password_style() {
     // `-P` is BSD only and not supported on GNU's `id`
     let username = whoami();
@@ -174,7 +170,6 @@ fn test_id_password_style() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_multiple_users() {
     unwrap_or_return!(check_coreutil_version(
         util_name!(),
@@ -232,7 +227,6 @@ fn test_id_multiple_users() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_multiple_users_non_existing() {
     unwrap_or_return!(check_coreutil_version(
         util_name!(),
@@ -300,16 +294,19 @@ fn test_id_multiple_users_non_existing() {
 }
 
 #[test]
-#[cfg(unix)]
+fn test_id_name_or_real_with_default_format() {
+    for flag in ["-n", "--name", "-r", "--real"] {
+        new_ucmd!()
+            .arg(flag)
+            .fails()
+            .stderr_only("id: printing only names or real IDs requires -u, -g, or -G\n");
+    }
+}
+
+#[test]
 fn test_id_default_format() {
     let ts = TestScenario::new(util_name!());
     for opt1 in ["--name", "--real"] {
-        // id: cannot print only names or real IDs in default format
-        let args = [opt1];
-        ts.ucmd()
-            .args(&args)
-            .fails()
-            .stderr_only(unwrap_or_return!(expected_result(&ts, &args)).stderr_str());
         for opt2 in ["--user", "--group", "--groups"] {
             // u/g/G n/r
             let args = [opt2, opt1];
@@ -337,22 +334,32 @@ fn test_id_default_format() {
 }
 
 #[test]
-#[cfg(unix)]
+fn test_id_zero_with_default_format() {
+    for z_flag in ["-z", "--zero"] {
+        new_ucmd!()
+            .arg(z_flag)
+            .fails()
+            .stderr_only("id: option --zero not permitted in default format\n");
+    }
+}
+
+#[test]
+fn test_id_zero_with_name_or_real() {
+    for z_flag in ["-z", "--zero"] {
+        for flag in ["-n", "--name", "-r", "--real"] {
+            new_ucmd!()
+                .args(&[z_flag, flag])
+                .fails()
+                .stderr_only("id: printing only names or real IDs requires -u, -g, or -G\n");
+        }
+    }
+}
+
+#[test]
 fn test_id_zero() {
     let ts = TestScenario::new(util_name!());
     for z_flag in ["-z", "--zero"] {
-        // id: option --zero not permitted in default format
-        ts.ucmd()
-            .args(&[z_flag])
-            .fails()
-            .stderr_only(unwrap_or_return!(expected_result(&ts, &[z_flag])).stderr_str());
         for opt1 in ["--name", "--real"] {
-            // id: cannot print only names or real IDs in default format
-            let args = [opt1, z_flag];
-            ts.ucmd()
-                .args(&args)
-                .fails()
-                .stderr_only(unwrap_or_return!(expected_result(&ts, &args)).stderr_str());
             for opt2 in ["--user", "--group", "--groups"] {
                 // u/g/G n/r z
                 let args = [opt2, z_flag, opt1];
@@ -378,9 +385,8 @@ fn test_id_zero() {
 #[test]
 #[cfg(feature = "feat_selinux")]
 fn test_id_context() {
-    use selinux::{self, KernelSupport};
-    if selinux::kernel_support() == KernelSupport::Unsupported {
-        println!("test skipped: Kernel has no support for SElinux context",);
+    if !uucore::selinux::is_selinux_enabled() {
+        println!("test skipped: Kernel has no support for SElinux context");
         return;
     }
     let ts = TestScenario::new(util_name!());
@@ -437,7 +443,6 @@ fn test_id_context() {
 }
 
 #[test]
-#[cfg(unix)]
 fn test_id_no_specified_user_posixly() {
     // gnu/tests/id/no-context.sh
 
@@ -453,18 +458,17 @@ fn test_id_no_specified_user_posixly() {
         feature = "feat_selinux"
     ))]
     {
-        use selinux::{self, KernelSupport};
-        if selinux::kernel_support() == KernelSupport::Unsupported {
-            println!("test skipped: Kernel has no support for SElinux context",);
-        } else {
+        if uucore::selinux::is_selinux_enabled() {
             let result = ts.ucmd().succeeds();
             assert!(result.stdout_str().contains("context="));
+        } else {
+            println!("test skipped: Kernel has no support for SElinux context");
         }
     }
 }
 
 #[test]
-#[cfg(all(unix, not(target_os = "android")))]
+#[cfg(not(target_os = "android"))]
 fn test_id_pretty_print_password_record() {
     // `-p` is BSD only and not supported on GNU's `id`.
     // `-P` is our own extension, and not supported by either GNU nor BSD.
@@ -475,3 +479,76 @@ fn test_id_pretty_print_password_record() {
         .fails()
         .stderr_contains("the argument '-p' cannot be used with '-P'");
 }
+
+#[test]
+#[cfg(all(feature = "chmod", feature = "chown"))]
+fn test_id_pretty_print_suid_binary() {
+    use uucore::process::{getgid, getuid};
+
+    if let Some(suid_coreutils_path) = create_root_owned_suid_coreutils_binary() {
+        let result = TestScenario::new(util_name!())
+            .cmd(suid_coreutils_path.to_str().unwrap())
+            .args(&[util_name!(), "-p"])
+            .succeeds();
+
+        // The `euid` line should be present only if the real UID does not belong to `root`
+        if getuid() == 0 {
+            result.stdout_does_not_contain("euid\t");
+        } else {
+            result.stdout_contains_line("euid\troot");
+        }
+
+        // The `rgid` line should be present only if the real GID does not belong to `root`
+        if getgid() == 0 {
+            result.stdout_does_not_contain("rgid\t");
+        } else {
+            result.stdout_contains("rgid\t");
+        }
+    } else {
+        print!("Test skipped; requires root user");
+    }
+}
+
+/// Create SUID temp file owned by `root:root` with the contents of the `coreutils` binary
+#[cfg(all(feature = "chmod", feature = "chown"))]
+fn create_root_owned_suid_coreutils_binary() -> Option<TempPath> {
+    use std::fs::read;
+    use std::io::Write;
+    use tempfile::NamedTempFile;
+    use uutests::util::{get_tests_binary, run_ucmd_as_root};
+
+    let mut temp_file = NamedTempFile::new().unwrap();
+    let coreutils_binary = read(get_tests_binary()).unwrap();
+    temp_file.write_all(&coreutils_binary).unwrap();
+    let temp_path = temp_file.into_temp_path();
+    let temp_path_str = temp_path.to_str().unwrap();
+
+    run_ucmd_as_root(&TestScenario::new("chown"), &["root:root", temp_path_str]).ok()?;
+    run_ucmd_as_root(&TestScenario::new("chmod"), &["+xs", temp_path_str]).ok()?;
+
+    Some(temp_path)
+}
+
+/// This test requires user with username 200 on system
+#[test]
+#[cfg(unix)]
+fn test_id_digital_username() {
+    match Command::new("id")
+        .arg("200")
+        .stdout(Stdio::null())
+        .stderr(Stdio::null())
+        .status()
+    {
+        Ok(ret) if ret.success() => {}
+        Ok(_) => {
+            println!("Test skipped; requires user with username 200 on system");
+            return;
+        }
+        Err(e) => {
+            println!("failed to run id command: {e}");
+            return;
+        }
+    }
+
+    new_ucmd!().arg("200").succeeds();
+}
